;;; 
;;; mymac-fn.el
;;;
;;; By Ross Ridge
;;; Public Domain
;;;

(defconst mymac-fn-ident "@(#) mymac mymac-fn.el 2.3 97/10/04 10:59:39")

(require 'byte-compile "bytecomp")

(defun mymac-fill-syntax-table (tbl last first value)
  ;;
  ;; Convienence function
  ;;
  (while (<= first last)
    (modify-syntax-entry first value tbl)
    (setq first (1+ first))))

(defvar mymac-syntax-table
  (let ((tbl (make-syntax-table)))
    (mymac-fill-syntax-table tbl 32 0 " ")
    (mymac-fill-syntax-table tbl ?@ ?! "w")
    (mymac-fill-syntax-table tbl ?Z ?A "w")
    (mymac-fill-syntax-table tbl ?` ?\[ "w")
    (mymac-fill-syntax-table tbl ?z ?a "w")
    (mymac-fill-syntax-table tbl ?~ ?{ "w")
    (mymac-fill-syntax-table tbl 255 127 " ")
    (modify-syntax-entry ?\` "\"" tbl)
    (modify-syntax-entry ?-  "_"  tbl)
    (modify-syntax-entry ?$  "'"  tbl)
    (modify-syntax-entry ?{  "(}" tbl)
    (modify-syntax-entry ?}  "){" tbl)
    (modify-syntax-entry ?\\ "\\" tbl)
    tbl))

(defvar mymac-follow-expansion nil)
(defvar mymac-obarray nil)

;;;
;;; There seems to be some sort of bug in the demacs release of GNU Emacs
;;; with `mapconcat'.  Redefining `mapconcat' below seems to fix the problem.
;;; 

(if (and (fboundp 'demacs-version) (not (fboundp 'oemacs-version)))
    (defun mapconcat (function sequence separator)
      (apply 'concat
	     (let ((list (mapcar function sequence))
		   (next))
	       (prog1
		   list
		 (while (setq next (cdr list))
		   (setcdr list (cons separator next))
		   (setq list next)))))))
	

(defun mymac-define-macro (macname &optional macval macfn bytecomp optimize)
  ;;
  ;; Define a mymac macro named MACNAME.
  ;; This function interns a symbol in the mymac obarray and
  ;; sets that symbol's value and function to MACVAL and MACFN
  ;; respectively.  If MACFN is nil then symbol `identity` is
  ;; used instead.  This function always returns nil.
  ;;
  (let ((macsym (intern macname mymac-obarray)))
    (set macsym macval)
    (if macfn
	(if bytecomp
	    (let ((macfnsym (make-symbol bytecomp))
		  (byte-optimize (not (not optimize))))
	      (fset macfnsym macfn)
	      (byte-compile macfnsym)
	      (fset macsym macfnsym))
	  (fset macsym macfn))
      (fset macsym 'identity))
    nil))

(defun mymac-slurp-expansion (string start end)
  ;;
  ;; Substitute and then expand the result of a macro expansion.
  ;; START and END mark the region to substitute STRING.
  ;; If STRING is nil then nothing is done and nil is returned.
  ;; If STRING isn't a string or the result of the expansion
  ;; is a empty list then t is returned.  Otherwise the
  ;; result of the expansion is returned as list of strings.
  ;;
  (if string
      (if (stringp string)
	  (progn (delete-region start end)
		 (goto-char start)
		 (insert-before-markers string)
		 (and mymac-follow-expansion
		      (or (pos-visible-in-window-p)
			  (recenter 0)))
		 (or (mymac-slurp-region start end)
		     t))
	t)))
  
(defun mymac-slurp-sexp ()
  ;;
  ;; Expand the next "sexp" after the point. 
  ;; This function returns nil if the end of the buffer is reached,
  ;; t if the expansion is empty, or the result of the expansion
  ;; as a list of strings.  The point is left at the end of the 
  ;; expanded sexp.
  ;; 
  (let ((end) (start) (schar) (macsym) (wasquote))
    (condition-case end-of-file
	(and (setq end (scan-sexps (point) 1))
	     (setq start (scan-sexps end -1))
	     (setq start (max (point) start)
	           end (set-marker (make-marker) (goto-char end))
	           schar (char-syntax (char-after start)))
	     (prog1
		 (cond
		  ((eq schar ?w )
		   (let ((macsym (intern-soft (buffer-substring start end)
					      mymac-obarray)))
		     (if (and macsym (fboundp macsym))
			 (mymac-slurp-expansion (funcall macsym
							 (symbol-value macsym))
						start end)
		       (buffer-substring start end))))
		  ((eq schar ?\( )
		   (if (and (/= (goto-char start)
				(progn (backward-prefix-chars)
				       (point)))
			    (goto-char start)
			    (looking-at "\\s([1-9][0-9]*\\s)"))
		       (string-to-int (buffer-substring (1+ start) (1- end)))
		     (if (eq (char-syntax (char-after (1+ start))) ?w)
			 (mymac-slurp-expansion (mymac-expand-macro (1+ start)
								    (1- end))
						start end)
		       (or (mymac-slurp-region (1+ start) (1- end))
			   t))))
		  ((eq schar ?\" ) 
		   (mymac-unquote-string (buffer-substring (1+ start)
							   (1- end))))
		  (t))
	       (goto-char end))))))

(defun mymac-slurp-region (start end) 
  ;;
  ;; Expand the region between START and END of the current buffer.
  ;; This function returns the result of the expansion as list of
  ;; strings.
  ;;
  (save-restriction
    (goto-char start)
    (narrow-to-region start end)
    (let ((sexp) (list))
      (while (eq (setq sexp (mymac-slurp-sexp)) t))
      (if (listp sexp)
	  (setq list sexp)
	(setq list (list sexp)))
      (while (setq sexp (mymac-slurp-sexp))
	(if (eq sexp t)
	    nil
	  (if (listp sexp)
	      (nconc list sexp)
	    (nconc list (list sexp)))))
      list)))

;;;
;;; no expansion -> nil
;;; null expansion -> t
;;; expansion -> string
;;;
    
(defun mymac-expand-macro (mstart mend)
  ;;
  ;; Expand the macro between MSTART and MEND. 
  ;; MSTART must point to the first character of the macro name.
  ;; If the macro isn't defined nil is returned.  The string the
  ;; macro is expands to is returned unless it expands to nothing
  ;; in which case t is returned.
  ;;
  (setq mend (set-marker (make-marker) mend))
  (let ((list) (macsym) (end))
    (condition-case end-of-file
	(and (setq end (scan-sexps mstart 1))
	     (setq macsym (intern-soft (buffer-substring mstart end)
				       mymac-obarray))
	     (fboundp macsym)
	     (or (apply macsym (symbol-value macsym) 
			  (prog1 (mymac-slurp-region end mend)
			    (if (> (point) mend)
				(debug))))
		 t)))))

      
;;(defun mymac-unquote-string (string)
;;  (let ((wasquote))
;;    (mapconcat (function (lambda (ch)
;;			   (setq wasquote
;;				 (and (eq ch mymac-quote-char)
;;				      (not wasquote)))
;;			   (if wasquote
;;			       ""
;;			     (char-to-string ch))))
;;	       string "")))

(defun mymac-unquote-string (string) 
  ;;
  ;; Remove one level of quote characters from the string STRING.
  ;;
  (let ((pos (string-match "\\s\\." string 0))
	(last 0)
	(newstr ""))
    (if (not pos)
	string
      (while (progn
	       (setq newstr (concat newstr (substring string last pos)))
	       (setq last (1+ pos))
	       (setq pos (string-match "\\s\\." string (1+ last)))))
      (concat newstr (substring string last)))))

(defun mymac-quote-string (string)
  ;;
  ;; Add one level of quotes to a string.
  ;;
  (concat "`" (mapconcat (function (lambda (arg)
				     (if (or (eq arg ?`)
					     (eq arg ?\\))
					 (concat "\\" (char-to-string arg))
				       (char-to-string arg))))
			 string "") "`"))

(defun mymac-expand-next-sexp (&optional trace)
  ;;
  ;; Expand the next sexp after the point.
  ;; This function returns nil if the end of the buffer is reached,
  ;; otherwise it returns t.
  ;;
  (condition-case end-of-file
      (let ((end) (start) (schar) (string) (macsym) (wasquote))
	(and (setq end (scan-sexps (point) 1))
	     (setq start (scan-sexps end -1))
	     (progn
	       (setq start (max (point) start)
		     end (set-marker (make-marker) (goto-char end))
		     schar (char-syntax (char-after start)))
	       (or
		(and (eq schar ?w)
		     (setq macsym 
			   (intern-soft (buffer-substring start end)
					mymac-obarray))
		     (fboundp macsym)
		     (progn
		       (delete-region start end)
		       (setq string (funcall macsym (symbol-value macsym)))))
		(and (eq schar ?\( )
		     (or (eq (char-syntax (char-after (1+ start))) ?w)
			 (progn
			   (if trace
			       (goto-char (1+ start))
			     (mymac-expand-region (1+ start) (1- end))
			     (goto-char end))
			   nil))
		     (setq string (mymac-expand-macro (1+ start) (1- end)))
		     (progn
		       (goto-char end)
		       (delete-region start end))))
	       (if (stringp string)
		   (progn (insert-before-markers string)
			  (and mymac-follow-expansion
			       (or (pos-visible-in-window-p)
				   (recenter 0)))
			  (if trace
			      (goto-char start)
			    (mymac-expand-region start end)
			    (goto-char end)))
		 (if (eq schar ?\")
		     (progn
		       (setq string (mymac-unquote-string
				     (buffer-substring (1+ start) (1- end))))
		       (delete-region start end)
		       (insert-before-markers string)
		       (and mymac-follow-expansion
			    (or (pos-visible-in-window-p)
				(recenter 0))))))
	       t)))))

(defun mymac-expand-region (start end)
  ;;
  ;; Expand all the macros in the region specified by START and END.
  ;;
  (save-restriction
    (goto-char start)
    (narrow-to-region start end)
    (while (mymac-expand-next-sexp))))

(defun mymac-new-obarray ()
  ;;
  ;; Returns an array suitable for use as an obarray for interning macro names.
  ;;
  (make-vector 251 0))

(defun mymac-obarray-to-list (in)
  ;;
  ;; Convert an obarray to a list of the symbols interned in it.
  ;;
  (let ((out))
    (mapatoms (function (lambda (sym)
			  (setq out (cons sym out))))
	      in)
    out))

(defun mymac-list-to-obarray (in)
  ;;
  ;; Intern all the symbols in a list into a new obarray.
  ;;
  (let ((out (mymac-new-obarray))
	(sym))
    (mapcar (function (lambda (osym)
			(setq sym (intern (symbol-name osym) out))
			(set sym (symbol-value osym))
			(fset sym (symbol-function osym))))
	    in)
    out))

(defun mymac-add-line-marks () 
  ;;
  ;; Add marks on each line returning them in an array.
  ;;
  (let* ((lines (count-lines 1 (1+ (buffer-size))))
	 (vec (make-vector (1+ lines) nil))
	 (i 0))
    (while (< i lines)
      (aset vec i (point-marker))
      (forward-line)
      (setq i (1+ i)))
    (aset vec i (buffer-size))
    vec))

(defun mymac-add-line-numbers (vec name)
  ;; 
  ;; Add #line directives where there were markers.
  ;;
  (let ((lines (1- (length vec)))
	(i 0)
	(last 1))
    (while (< i lines)
      (while (progn
	       (goto-char (aref vec i))
	       (setq i (1+ i))
	       (= (point) (aref vec i))))
      (and (/= (count-lines last (point)) 1)
	   (insert-before-markers (if (bolp) "" "\n") 
				  "#line " (int-to-string i) 
				  " \"" name "\"\n"))
      (setq last (point)))))
      
(defun mymac-expand-buffer (&optional src-buffer dest-buffer filename)
  "Expand all the mymac macros in a buffer.
By default this function expands the current buffer and creates a new
buffer to store the expansion.  The optional arguments SRC-BUFFER and
DEST-BUFFER can be used to explicity specify which buffers to use."
  (interactive)
  (save-excursion
    (setq src-buffer (or src-buffer (current-buffer)))
    (if (eq src-buffer dest-buffer)
	(set-buffer dest-buffer)
      (let ((args (list (point-min) (point-max))))
	(set-buffer (or dest-buffer
			(get-buffer-create (concat "*expansion of "
						   (buffer-name src-buffer)
						   "*"))))
	(delete-region (point-min) (point-max))
	(apply 'insert-buffer-substring src-buffer args)))
    (if mymac-follow-expansion
	(pop-to-buffer (current-buffer))
      (display-buffer (current-buffer)))
    (make-local-variable 'mymac-obarray)
    (setq mymac-obarray (mymac-list-to-obarray mymac-macro-list))
    (set-syntax-table mymac-syntax-table)
    (if filename
	(mymac-add-line-numbers (prog1 (mymac-add-line-marks)
				  (mymac-expand-region (point-min)
						       (point-max)))
				filename)
      (mymac-expand-region (point-min) (point-max)))
    (display-buffer (current-buffer))))

(defun mymac-expand-buffer-in-place (&optional buf)
  "Expand the macros in the current buffer in place.
The optional argument BUF can be used to specify a buffer other
than the current buffer to expand."
  (interactive)
  (setq buf (or buf (current-buffer)))
  (mymac-expand-buffer buf buf))

(defun mymac-expand-file (src dest &optional lineno)
  "Expand the macros in the file SRC and write expansion to the file DEST."
  (interactive "fFile to expand: \nFExpand %s to: ")
  (find-file dest)
  (delete-region (point-min) (point-max))
  (insert-file-contents src)
  (mymac-expand-buffer (current-buffer) (current-buffer) (and lineno src))
  (save-buffer))

(defun mymac-batch-expand-files (&optional lineno)
  ;;
  ;; In batch mode, expand the files specified on command line.  The
  ;; command line arguments are source and destination filename pairs.
  ;;
  (if (or noninteractive debug-on-error)
      (let ((src) (dest))
	(while (and (setq src (car command-line-args-left))
		    (setq dest (car (cdr command-line-args-left))))
	  (message "Expanding %s to %s..." src dest)
	  (mymac-expand-file src dest lineno)
	  (setq command-line-args-left (cdr (cdr command-line-args-left))))
	(message "Done")
	(kill-emacs 0))
    (error "mymac-batch-file-expand may only be called in batch mode")))

(defun mymac-batch-expand-files-lineno ()
  (mymac-batch-expand-files t))

(defun mymac-print-macro (sym)
  ;;
  ;; Print the definition of macro who's symbol is SYM.
  ;;
  (print (list sym
	       (if (boundp sym)
		   (symbol-value sym)
		 'void)
	       (if (fboundp sym)
		   (symbol-function sym)
		 'void))))

(defun mymac-dump-macros ()
  "Dump a list of all the mymac macros currently defined."
  (interactive)
    (if mymac-obarray
	(with-output-to-temp-buffer "*mymac local macros*"
	  (mapatoms 'mymac-print-macro mymac-obarray))
      (with-output-to-temp-buffer "*mymac global macros*"
	(mapcar 'mymac-print-macro mymac-macro-list))))

(defun mymac-define-expand (value &rest args)
  ;;
  ;; Expand an ordinary macro definition.
  ;;
  (mapconcat (function (lambda (arg)
			 (if (integerp arg)
			     (nth (1- arg) args)
			   arg)))
	     value ""))

(defun mymac-search-path (filename path)
  ;;
  ;; Search PATH (a list of directories) for FILENAME.
  ;;
  (let ((tmp (car path)))
    (setq tmp (if tmp
		  (concat tmp "/" filename)
		filename))
    (if (file-exists-p tmp)
	tmp
      (and path
	   (mymac-search-path filename (cdr path))))))

(defun mymac-load-default-macros () 
  ;;
  ;; Load the default macros from the file mymacdef.mm
  ;;
  (let ((mymac-obarray (mymac-new-obarray)))
    (mymac-define-macro "bind-lisp-expr" nil    
			'(lambda (value macname lisp-expr
					&optional bytecomp opt)
			   (let ((funct (eval (read lisp-expr))))
			     (or (and (symbolp funct) (fboundp funct))
				 (eq (car funct) 'lambda)
				 (setq funct 
				       (cons 'lambda 
					     (cons '(value &rest args)
						   (cons funct nil)))))
			     (mymac-define-macro macname value funct
						 bytecomp opt)))
			"*bind-lisp-expr*" t)
    (message "Reading mymacdef.mm...")
    (save-excursion
      (set-buffer (generate-new-buffer "*mymac startup*"))
      (insert-file-contents (mymac-search-path "mymacdef.mm" load-path))
      (set-syntax-table mymac-syntax-table)
      (mymac-slurp-region (point-min) (point-max))
      (kill-buffer (current-buffer)))
    (message "Reading mymacdef.mm... done.")
    (mymac-obarray-to-list mymac-obarray)))

(provide 'mymac-fn)







